/////////////////////////////////  \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\  /////////////////////////////////
//////////////////////// Crispy.SpectateLib \\\\\\\\\\\\\\\\\\\\\\\\
//  Created by Crispy. Send problems, questions, and suggestions  \\
//  to cpelling@webone.com.au; flames to /dev/null :-)            \\
/////////////////////////////////  \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\  /////////////////////////////////

/*

Crispy.SpectateLib allows you to easily put spectators into the game.
Spectators cannot use verbs attached to other objects, or click on
anything that is not a screen object. You can also specify objects
they are always allowed to click on, and objects they are never allowed
to click on.

----------------------------------------------------------------------------
 IMPORTANT NOTE: A BYOND bug causes spectators to sometimes have access
 to verbs attached to other objects. You should include a check in all
 of your non-spectator verbs, as follows:

	if (usr.isspectator()) return

 Including that check will prevent spectators from using verbs they
 shouldn't have access to (for example, a get verb). This is a temporary
 solution; the bug will hopefully be fixed in a future BYOND version.
----------------------------------------------------------------------------

It is highly recommended that ordinary players are defined as a subtype of
/mob (such as /mob/player), to prevent spectators gaining verbs they are
not supposed to have.

To make someone into a spectator, simply create a spectator mob
(/mob/cr_watch) and assign the client.mob var to it:

mob/player/verb/spectate()
	client.mob=new/mob/cr_watch()
	del src //Delete old mob

You can customize your spectators by modifying vars belonging
to /mob/cr_watch; user-modifiable vars are as follows:

	canobserve    - If true (default), spectators are given the observe verb.
	                Set this to false to turn the observe verb into a proc;
	                you might want to do this if you want to use a HUD, or
	                customize the name/description of the verb.

	loginmsg      - Change this to customize the message spectators get when
	                they log in. Set it to null to disable this message.

	nomapedge     - If false (default), spectators will be able to see the
	                black space over the edge of the map. If true, they will
	                not be able to move close enough to see the black.

	startx,       - These vars set the location of the free-roaming eye.
	starty,         By default, they are all 1 - you will probably want to
	startz          change this to something more meaningful.

	allowclick[]  - Empty by default. This is a list of the types of objects
	                that spectators are always allowed to click on. Subtypes
	                count, as well - if /obj/myobj is in this list, the player
	                may not only click on objs of type /obj/myobj, but
	                /obj/myobj/anotherkindofobj as well.

	denyclick[]   - Empty by default. This is a list of the types of objects
	                that spectators are never allowed to click on. Subtypes are
	                treated the same way as with allowclick[].

SpectateLib also provides the following global procs:

	cr_getviewx(viewsize)
		When provided with a view size in a format accepted by client.view or
		world.view, returns the width  (or x distance) from the centre of the
		view to the edge of the map screen. That is, if passed 5 it will return 5.
		If passed the string "11x11", it will also return 5. If passed "9x7", it
		will return 4.

	cr_getviewy(viewsize)
		Similar to cr_getviewx(), but returns the height (or y) portion of the
		view size. If passed 5 or "11x11", it will return 5. If passed "9x7", it
		will return 3.


Copyright 2003, Crispy.

*/


mob/cr_watch //Spectators
	density=0
	icon=null
	see_in_dark=10
	see_invisible=100
	luminosity=0
	//Can see everything
	sight = (SEE_MOBS|SEE_OBJS|SEE_TURFS)
	//If true, free-roaming. If false, observing.
	var/free=1
	//Current free-roaming location
	var/tmp/atom/roamingloc

	////User-modifiable settings

	//If true, the observe command is available as a verb.
	var/canobserve=1

	//Message displayed to spectators on login
	var/loginmsg="You are a spectator. You may roam freely using the arrow \
					keys, or use the observe command \
					to watch another player."

	//If true, stops the spectator from going close enough to the
	//edge of the map to see the black.
	var/nomapedge=0

	//Clicking on objects
	var/allowclick[0] //List of types the spectator is always allowed to click on.
	var/denyclick[0] //List of types spectator is never allowed to click on.

	//These vars determine where to start the eye
	var/startx=1
	var/starty=1
	var/startz=1

	Login()
		..()
		//Put the mob at null
		loc=null
		//Initialise eye
		roamingloc=locate(startx,starty,startz)
		client.eye=roamingloc
		client.perspective=EYE_PERSPECTIVE
		//Initialise player
		name=key
		gender=client.gender
		//Observe verb?
		if (canobserve) verbs+=/mob/cr_watch/proc/observe
		//Login message
		if (loginmsg) src << loginmsg

	Move(atom/NewLoc,MoveDir)
		////Move the eye
		//Find location
		var/atom/A=get_step(client.eye,MoveDir)

		//Keep eye within boundaries
		if (A && nomapedge)
			var/x=A.x
			var/y=A.y

			////Get view size
			var/viewx
			var/viewy
			if (client)
				//Use client.view if possible
				viewx=cr_getviewx(client.view)
				viewy=cr_getviewy(client.view)
			else
				//Otherwise use world.view
				viewx=cr_getviewx(world.view)
				viewy=cr_getviewy(world.view)

			if (x <= viewx) x=viewx+1
			else if (x > world.maxx-viewx) x=world.maxx-viewx
			if (y <= viewy) y=viewy+1
			else if (y > world.maxy-viewy) y=world.maxy-viewy
			A=locate(x,y,A.z)

		//Move to location
		if (A)
			client.eye=A
			roamingloc=A

			if (!free)
				free=1
				src << "You return to free roaming mode"

	proc/observe()
		//Observe other players
		set desc="Toggle observe mode. In observe mode, you can watch other players."
		if (free)
			var/client/C=input(src,"Choose a player to observe.","Observe") as null|anything in cr_clientlist
			if (!C || !C.mob) return
			client.eye=C.mob
			free=0
			src << "You start observing [C.key]"
		else
			client.eye=roamingloc
			free=1
			src << "You return to free roaming mode"

//Proc to check if a mob is of the type /mob/cr_watch
mob/proc/isspectator()
	return 0

mob/cr_watch/isspectator()
	return 1

//Find out the x and y view sizes, given a value that is either a number or a text string in the form
//of WIDTHxHEIGHT (which can be used with world.view or client.view)
proc/cr_getviewx(view)
	if(isnum(view))
		return view
	else if (istext(view))
		return round(text2num(copytext(view,1,findtext(view,"x")))/2)

proc/cr_getviewy(view)
	if(isnum(view))
		return view
	else if (istext(view))
		return round(text2num(copytext(view,findtext(view,"x")+1))/2)


client/Click(Object,Location)
	if (allowedtoclick(Object))
		//Allowed; proceed as usual
		.=..()

client/DblClick(Object,Location)
	if (allowedtoclick(Object))
		//Allowed; proceed as usual
		.=..()

client/proc/allowedtoclick(O)
	.=1 //Assume we're allowed
	if (istype(mob,/mob/cr_watch))
		//Spectator; deny some (if not all) clicks

		var/mob/cr_watch/M=mob

		//If Object is an atom that isn't a HUD object
		if (O && istype(O,/atom) && !(O in screen))
			//Test to see if it is always allowed
			for (var/sometype in M.allowclick)
				if (istype(O,sometype))
					return 1
			//Otherwise, deny it
			return 0
		else
			//Test to see if it is always denied
			for (var/sometype in M.denyclick)
				if (istype(O,sometype))
					return 0

//Manage a list of clients (for the observe verb)
var/tmp/client/cr_clientlist[0] //List of clients

client
	pixel_step_size=32 //Stop the view sliding all over the place
	New()
		cr_clientlist+=src
		.=..()
	Del()
		cr_clientlist-=src
		.=..()